use anyhow::Result;
use asset::*;
use assetpath::AssetPath;

use crate::prelude::*;

use std::collections::HashMap;
use std::sync::{
    Arc, RwLock, RwLockReadGuard,
    atomic::{AtomicU32, Ordering},
};
use std::thread;

#[derive(Debug)]
pub struct AssetCache {
    // asset data
    assets: Arc<RwLock<HashMap<String, Asset>>>,

    // thread safe ready counter
    ready_count: Arc<AtomicU32>,

    // value to compare against
    total_count: u32,
}

impl AssetCache {
    pub(crate) fn new(
        context: &Context,
        engine_settings: &EngineSettings,
        loaded_assets: impl IntoIterator<Item = (String, Asset)>,
        names: impl IntoIterator<Item = (AssetPath, String)>,
    ) -> Result<Self> {
        // init vars
        let mut total_count = 0;
        let ready_count = Arc::new(AtomicU32::new(0));
        let mut assets = HashMap::new();

        // add already loaded assets into map
        for (name, asset) in loaded_assets.into_iter() {
            assets.insert(name, asset);

            total_count += 1;
            ready_count.fetch_add(1, Ordering::SeqCst);
        }

        // create a rw lock around asset map
        let rw_lock_assets = Arc::new(RwLock::new(assets));

        // start threads for every asset still to load
        for (path, name) in names.into_iter() {
            total_count += 1;

            let device = context.device().clone();
            let queue = context.queue().clone();
            let render_type = engine_settings.graphics_info()?.render_type;

            let rw_lock_assets = rw_lock_assets.clone();
            let ready_count = ready_count.clone();

            thread::spawn(move || -> Result<()> {
                if cfg!(debug_assertions) {
                    println!("AssetCache: started thread for gltf file: {}", &path);
                }

                let gltf_asset = GltfAsset::new(&path)?;
                let asset = Asset::new(&device, &queue, render_type, gltf_asset, &name)?;

                {
                    let mut locked_assets = rw_lock_assets.write().unwrap();
                    locked_assets.insert(name.clone(), asset);
                }

                ready_count.fetch_add(1, Ordering::SeqCst);

                if cfg!(debug_assertions) {
                    println!("AssetCache: finished loading gltf file: {}", &path);
                }

                Ok(())
            });
        }

        Ok(AssetCache {
            assets: rw_lock_assets,
            ready_count,
            total_count,
        })
    }

    pub(crate) fn get_all(&self) -> Result<RwLockReadGuard<'_, HashMap<String, Asset>>> {
        Ok(self.assets.read().unwrap())
    }

    /// Returns all names of assets stored in this cache
    pub fn assets(&self) -> Result<Vec<String>> {
        assert!(
            self.finished(),
            "AssetCache is not ready loading its assets"
        );

        Ok(self.assets.read().unwrap().keys().cloned().collect())
    }

    pub fn finished(&self) -> bool {
        self.ready_count.load(Ordering::SeqCst) == self.total_count
    }

    pub fn currently_loaded(&self) -> u32 {
        self.ready_count.load(Ordering::SeqCst)
    }

    /// Merges another AssetCache into this
    pub fn merge(&mut self, mut _other: AssetCache) {
        todo!()

        // for (key, value) in other.assets.iter_mut() {
        //     if self.assets.get(key).is_none() {
        //         self.assets.insert(key.clone(), value.clone());
        //     }
        // }
    }
}
