use utilities::prelude::cgmath::{vec3, InnerSpace, Matrix4, Vector3};

#[derive(Debug)]
pub struct ViewFrustum {
    near_left_top: Vector3<f32>,
    near_left_bottom: Vector3<f32>,
    near_right_top: Vector3<f32>,
    near_right_bottom: Vector3<f32>,
    far_left_top: Vector3<f32>,
    far_left_bottom: Vector3<f32>,
    far_right_top: Vector3<f32>,
    far_right_bottom: Vector3<f32>,
}

impl ViewFrustum {
    pub(crate) fn new(inv_mvp: Matrix4<f32>) -> Self {
        let near_left_top = vec3(-1.0, -1.0, -1.0);
        let near_left_bottom = vec3(-1.0, 1.0, -1.0);
        let near_right_top = vec3(1.0, -1.0, -1.0);
        let near_right_bottom = vec3(1.0, 1.0, -1.0);
        let far_left_top = vec3(-1.0, -1.0, 1.0);
        let far_left_bottom = vec3(-1.0, 1.0, 1.0);
        let far_right_top = vec3(1.0, -1.0, 1.0);
        let far_right_bottom = vec3(1.0, 1.0, 1.0);

        Self {
            near_left_top: Self::to_world(inv_mvp, near_left_top),
            near_left_bottom: Self::to_world(inv_mvp, near_left_bottom),
            near_right_top: Self::to_world(inv_mvp, near_right_top),
            near_right_bottom: Self::to_world(inv_mvp, near_right_bottom),
            far_left_top: Self::to_world(inv_mvp, far_left_top),
            far_left_bottom: Self::to_world(inv_mvp, far_left_bottom),
            far_right_top: Self::to_world(inv_mvp, far_right_top),
            far_right_bottom: Self::to_world(inv_mvp, far_right_bottom),
        }
    }

    fn to_world(inv_mvp: Matrix4<f32>, p: Vector3<f32>) -> Vector3<f32> {
        let c = inv_mvp * p.extend(1.0);

        vec3(c.x / c.w, c.y / c.w, c.z / c.w)
    }

    pub fn planes(&self) -> FrustumPlanes {
        FrustumPlanes {
            left: Plane::new(self.near_left_bottom, self.near_left_top, self.far_left_top),
            right: Plane::new(
                self.near_right_top,
                self.near_right_bottom,
                self.far_right_top,
            ),
            top: Plane::new(self.near_left_top, self.near_right_top, self.far_left_top),
            bottom: Plane::new(
                self.near_right_bottom,
                self.near_left_bottom,
                self.far_left_bottom,
            ),

            near: Plane::new(
                self.near_left_bottom,
                self.near_right_bottom,
                self.near_right_top,
            ),
            far: Plane::new(
                self.far_right_bottom,
                self.far_left_bottom,
                self.far_right_top,
            ),
        }
    }
}

#[derive(Debug)]
pub struct FrustumPlanes {
    pub left: Plane,
    pub right: Plane,
    pub top: Plane,
    pub bottom: Plane,

    pub near: Plane,
    pub far: Plane,
}

impl FrustumPlanes {
    pub fn as_array(&self) -> [&Plane; 6] {
        [
            &self.left,
            &self.right,
            &self.top,
            &self.bottom,
            &self.near,
            &self.far,
        ]
    }
}

#[derive(Debug)]
pub struct Plane {
    pub point: Vector3<f32>,
    pub normal: Vector3<f32>,
}

impl Plane {
    pub fn new(p1: Vector3<f32>, p2: Vector3<f32>, p3: Vector3<f32>) -> Self {
        let v1 = (p1 - p2).normalize();
        let v2 = (p1 - p3).normalize();

        let n = v1.cross(v2).normalize();

        Self {
            point: p1,
            normal: n,
        }
    }

    pub fn is_above(&self, p: Vector3<f32>) -> bool {
        self.normal.dot(p - self.point) > 0.0
    }
}

#[test]
fn plane_test() {
    let n = vec3(-1.0, 0.0, 0.0);
    let p = vec3(0.0, 0.0, 0.0);

    let plane = Plane {
        point: p,
        normal: n,
    };

    let q1 = vec3(2.0, 1.0, 0.0);
    let q2 = vec3(-3.0, 3.5, 0.0);

    assert!(!plane.is_above(q1));
    assert!(plane.is_above(q2));
}
